/*
 * Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "sheep_priv.h"

static struct sheepdog_config config;

char *config_path, *node_config_path;

#define CONFIG_PATH "/config"

static int write_config(void)
{
	int ret;

	ret = atomic_create_and_write(config_path, (char *)&config,
				      sizeof(config), true, false);
	if (ret < 0) {
		sd_err("atomic_create_and_write() failed");
		return SD_RES_EIO;
	}

	return SD_RES_SUCCESS;
}

static void check_tmp_config(void)
{
	int ret;
	char tmp_config_path[PATH_MAX];

	snprintf(tmp_config_path, PATH_MAX, "%s.tmp", config_path);

	ret = unlink(tmp_config_path);
	if (!ret || ret != ENOENT)
		return;

	sd_info("removed temporal config file");
}

static int get_cluster_config(struct cluster_info *cinfo)
{
	cinfo->ctime = config.ctime;
	cinfo->nr_copies = config.copies;
	if (config.ctime > 0)
		cinfo->flags = config.flags;
	else
		cinfo->flags = (config.flags & ~SD_CLUSTER_FLAG_AUTO_VNODES) |
			(cinfo->flags & SD_CLUSTER_FLAG_AUTO_VNODES);
	cinfo->copy_policy = config.copy_policy;
	cinfo->block_size_shift = config.block_size_shift;
	memcpy(cinfo->default_store, config.default_store,
	       sizeof(config.default_store));

	return SD_RES_SUCCESS;
}

int init_node_config_file(void)
{
	int fd, ret;
	struct stat st;

	ret = stat(node_config_path, &st);
	if (ret < 0)
		/* this node doesn't have a node config, do nothing */
		return 0;

	fd = open(node_config_path, O_RDONLY);
	if (fd < 0) {
		sd_err("failed to open node config (%s): %m", node_config_path);
		return -1;
	}

	ret = xread(fd, &sys->ninfo, sizeof(sys->ninfo));
	if (ret != sizeof(sys->ninfo)) {
		sd_err("failed to read node config (%s): %m", node_config_path);
		return -1;
	}

	close(fd);
	return 0;
}

int init_config_file(void)
{
	int fd, ret = 0;

	check_tmp_config();

	fd = open(config_path, O_RDONLY);
	if (fd < 0) {
		if (errno != ENOENT) {
			sd_err("failed to read config file, %m");
			return -1;
		}
		goto create;
	}

	ret = xread(fd, &config, sizeof(config));
	if (ret == 0) {
		close(fd);
		goto create;
	}
	if (ret < 0) {
		sd_err("failed to read config file, %m");
		goto out;
	}

	if (config.version != SD_FORMAT_VERSION) {
		sd_err("This sheep version is not compatible with"
		       " the existing data layout, %d", config.version);
		if (sys->upgrade) {
			/* upgrade sheep store */
			ret = sd_migrate_store(config.version, SD_FORMAT_VERSION);
			if (ret == 0) {
				/* reload config file */
				ret = xpread(fd, &config, sizeof(config), 0);
				if (ret != sizeof(config)) {
					sd_err("failed to reload config file,"
					       " %m");
					ret = -1;
				} else {
					ret = 0;
					goto reload;
				}
			}
			goto out;
		}

		sd_err("use '-u' option to upgrade sheep store");
		ret = -1;
		goto out;
	}

reload:
	if ((config.flags & SD_CLUSTER_FLAG_AUTO_VNODES) !=
			(sys->cinfo.flags & SD_CLUSTER_FLAG_AUTO_VNODES)
		&& !sys->gateway_only
		&& config.ctime > 0) {
		sd_err("Designation of before a restart and a vnodes option is different.");
		return -1;
	}

	ret = 0;
	get_cluster_config(&sys->cinfo);
	if ((config.flags & SD_CLUSTER_FLAG_DISKMODE) !=
	    (sys->cinfo.flags & SD_CLUSTER_FLAG_DISKMODE)) {
		sd_err("This sheep can't run because "
		       "exists data format mismatch");
		return -1;
	}

create:
	config.version = SD_FORMAT_VERSION;
	if (write_config() != SD_RES_SUCCESS)
		return -1;

out:
	close(fd);

	return ret;
}

void init_config_path(const char *base_path)
{
	int len = strlen(base_path) + strlen(CONFIG_PATH) + 1;

	config_path = xzalloc(len);
	snprintf(config_path, len, "%s" CONFIG_PATH, base_path);

	len = strlen(base_path) + strlen(NODE_CONFIG_PATH) + 1;
	node_config_path = xzalloc(len);
	snprintf(node_config_path, len, "%s" NODE_CONFIG_PATH, base_path);
}

int set_cluster_config(const struct cluster_info *cinfo)
{
	config.ctime = cinfo->ctime;
	config.copies = cinfo->nr_copies;
	config.copy_policy = cinfo->copy_policy;
	config.flags = cinfo->flags;
	config.block_size_shift = cinfo->block_size_shift;
	memset(config.default_store, 0, sizeof(config.default_store));
	pstrcpy((char *)config.default_store, sizeof(config.default_store),
		(char *)cinfo->default_store);

	return write_config();
}

int set_node_space(uint64_t space)
{
	config.space = space;

	return write_config();
}

int get_node_space(uint64_t *space)
{
	*space = config.space;

	return SD_RES_SUCCESS;
}

bool is_cluster_formatted(void)
{
	struct cluster_info cinfo;

	get_cluster_config(&cinfo);

	return cinfo.ctime != 0;
}

int set_cluster_shutdown(bool down)
{
	config.shutdown = down;
	return write_config();
}

bool was_cluster_shutdowned(void)
{
	return config.shutdown;
}

static inline __attribute__((used)) void __sd_config_format_build_bug_ons(void)
{
	/* never called, only for checking BUILD_BUG_ON()s */
	BUILD_BUG_ON(sizeof(struct sheepdog_config) != SD_CONFIG_SIZE);
}
